#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Joy Cheung <joycheung1994@gmail.com>
    Yalin Li

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


from sympy import symbols, sympify, simplify, Matrix, solve
from sympy.parsing.sympy_parser import parse_expr
import numpy as np

__all__ = ('get_stoichiometric_coeff', )

#%%

def split_coefficient(nID, sign):
    if len(nID.split(']')) > 1:
        coeff, ID = nID.split(']')[0].split('[')[1], nID.split(']')[1]
        try: return sign * float(coeff), ID
        except ValueError: 
            if sign == 1: return coeff, ID
            elif sign == -1: return '-('+coeff+')', ID
    else:
        for i, letter in enumerate(nID):
            if letter != 'e' and letter.isalpha(): break
        if i: 
            ID = nID[i:]
            n = sign * float(nID[:i])
        else: 
            ID = nID
            n = sign
    return n, ID

def extract_coefficients(nIDs, dct, sign):
    for nID in nIDs:
        n, ID = split_coefficient(nID, sign)
        if ID in dct:
            raise ValueError('Components can only appear once in a reaction; '
                            f'multiple instances of {repr(ID)} found')
        dct[ID] = n        

def str2dct(reaction) -> dict:
    reaction = reaction.replace(' ', '')
    left, right = reaction.split('->')
    reactants = left.split('+')
    products = right.split('+')
    dct = {}
    extract_coefficients(reactants, dct, -1)
    extract_coefficients(products, dct, 1)
    return dct

def dct2arr(dct, components):
    arr = np.zeros(components.size)
    cmp_index = components._index
    for ID, coefficient in dct.items():
        arr[cmp_index[ID]] = coefficient
    return arr

def dct2list(dct, components):
    lst = [0] * components.size
    cmp_index = components._index
    for ID, coefficient in dct.items():
        lst[cmp_index[ID]] = coefficient
    return lst    

def get_ic(cmps, conservation_for):        
    '''
    return conversion factors as a sympy matrix to solve for unknown stoichiometric coefficients.
    '''
    if conservation_for:
        arr = getattr(cmps, 'i_'+conservation_for[0])
        for c in conservation_for[1:]:
            arr = np.vstack((arr, getattr(cmps, 'i_'+c)))
        return Matrix(arr.tolist())
    else: return None

def symbolize(coeff_dct, components, conserved_for, parameters):
    n = sum([v in ('?', '-(?)') for v in coeff_dct.values()])
    if n > 0:
        unknowns = symbols('unknown0:%s' % n)
        i = 0
        v_arr = []
        IDs = sorted(coeff_dct)
        for cmp in IDs:
            if coeff_dct[cmp] in ('?', '-(?)'):
                v_arr.append('unknown%s' % i)
                i += 1
            else: v_arr.append(coeff_dct[cmp])
        v = Matrix(sympify(v_arr, parameters))
        ic = get_ic(components.subgroup(IDs), conserved_for)
        sol = solve(ic * v, unknowns)
        coeff_dct = dict(zip(IDs, simplify(v.subs(sol))))
        del unknowns                    
    else:
        coeff_dct = {k: simplify(parse_expr(v, local_dict=parameters)) for k, v in coeff_dct.items()}
    return coeff_dct
    
def get_stoichiometric_coeff(reaction, ref_component, components, conserved_for, parameters):
    '''
    parse input reaction to get array of symbolic expressions (or a function to return numpy array with parameters as kwargs) 
    or numpy array of input values for stoichiometric coefficients.
    '''
    isa = isinstance
    coeff_dct = None    

    if isa(reaction, np.ndarray):
        if len(reaction) == components.size: coeff = reaction
        else: 
            raise ValueError("The array of stoichiometric coefficients must have "
                             "the same length as Components {repr(components))}.")
    elif isa(reaction, dict):
        if all([isa(v, (float, int)) for v in reaction.values()]):
            coeff = dct2arr(reaction, components)
        else: 
            coeff_dct = reaction
    elif isa(reaction, str):
        coeff_dct = str2dct(reaction)
    else:
        raise ValueError("reaction must be either a str or a dict or an array; "
                         f"not a '{type(reaction).__name__}' object")        
    if coeff_dct:
        coeff_dct = symbolize(coeff_dct, components, conserved_for, parameters)
        coeff = dct2list(coeff_dct, components)
    if ref_component:
        normalize_factor = abs(coeff[components._index[ref_component]])
        if isa(coeff, np.ndarray): coeff /= normalize_factor
        else: coeff[:] = [v/normalize_factor for v in coeff]
    return coeff

#%%
# from qsdsan import Components

# # hydrolysis
# rxn = 'XB_Subst -> [1-fsi]SF + [fsi]SU_E + [?]SNH4 + [?]SPO4 + [?]SCO3'
# ref = 'XB_Subst'
# cmps = Components.load_default()
# conserved_for = ('COD', 'N', 'P', 'charge')
# parameters = {'fsi': symbols('fsi')}

# # anoxic growth on SF
# rxn = '[1/yh]SF + [(1-yh)/yh/2.86]SNO3 + [?]SPO4 + [?]SNH4 + [?]SCO3 -> [(1-yh)/yh/2.86]SN2 + XOHO'
# ref = 'XOHO'
# # cmps = Components.load_default()
# # conserved_for = ('COD', 'N', 'P', 'charge')
# parameters = {'yh': symbols('yh'),}

# stoichiometry = get_stoichiometric_coeff(rxn, ref, cmps, conserved_for, parameters)
# dict(zip(cmps.IDs, stoichiometry))

#%%
# nID = '[1-fsi]SF'
# coeffient = '1-fsi'

# split_coefficient(nID, 1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

import pandas as pd
from thermosteam.utils import registered
from . import currency, ImpactItem
from .utils import (
    auom, copy_attr,
    format_number as f_num
    )

__all__ = ('Transportation',)


@registered(ticket_name='Trans')
class Transportation:
    '''    
    Transportation activity for cost and environmental impact calculations.
    
    Parameters
    ----------
    ID : str
        ID of this transportation activity.
    item : :class:`ImpactItem`
        Impact item associated with this transportation activity.
    load_type : str
        Can be either 'mass' or 'volume'.
    load : float
        Quantity of the load per trip.
    load_unit : str
        Unit of the load.
    distance : float
        Distance per trip.
    distance_unit : str
        Unit of the distance.
    interval : float
        Distance per trip.
    interval_unit : str
        Unit of the transportation interval.
    
    Examples
    --------
    >>> import qsdsan as qs
    >>> # Make impact indicator
    >>> GWP = qs.ImpactIndicator('GlobalWarming', alias='GWP', unit='kg CO2-eq')
    >>> FEC = qs.ImpactIndicator('FossilEnergyConsumption', alias='FEC', unit='MJ')
    >>> # Assuming transporting 1 kg of goods for 1 km emits 10 kg CO2-eq
    >>> Trucking = qs.ImpactItem('Trucking', 'kg*km', GWP=10, FEC=5)
    >>> # Make a transportation activity for transporting 1000 kg goods for 1 mile every day
    >>> shipping = qs.Transportation('shipping', item=Trucking,
    ...                              load_type='mass', load=1, load_unit='tonne',
    ...                              distance='1', distance_unit='mile',
    ...                              interval='1', interval_unit='day')
    >>> shipping.show()
    Transportation: shipping
    Impact item   : Trucking [per trip]
    Load          : 1000 kg
    Distance      : 1.61 km
    Interval      : 24 hr
    Total cost    : None USD
    Total impacts :
                                  Impacts
    GlobalWarming (kg CO2-eq)    1.61e+04
    FossilEnergyConsumption (MJ) 8.05e+03
    '''

    __slots__ = ('_ID', '_item', '_load_type', '_load', '_distance', '_interval',
                 'default_units')
    
    def __init__(self, ID='', item=None,
                 load_type='mass', load=1., load_unit='kg',
                 distance=1., distance_unit='km',
                 interval=1., interval_unit='hr'):
        self._register(ID)
        self.item = item
        self.default_units = {
            'distance': 'km',
            'interval': 'hr',
            }
        
        self._load_type = load_type
        if load_type == 'mass':
            self.default_units['load'] = 'kg'
            self.default_units['quantity'] = 'kg*km'
        elif load_type == 'volume':
            self.default_units['load'] = 'm3'
            self.default_units['quantity'] = 'm3*km'
        else:
            raise ValueError("load_type can only be 'mass' or 'volume', "
                             f'not {load_type}.')
        
        self._update_value('load', load, load_unit)
        self._update_value('distance', distance, distance_unit)
        self._update_value('interval', interval, interval_unit)

        if item:            
            try:
                auom(str(load_unit)+'*'+str(distance_unit)).convert(1, self.item.functional_unit)
            except:
                raise ValueError(f'Units of `load` {load_unit} and `distance` {distance_unit} '
                                 f'do not match the item `functional_unit` {self.item.functional_unit}.')
        
    
    def _update_value(self, var, value, unit=''):
        default_unit = self.default_units[var]
        if not unit or unit == default_unit:
            setattr(self, '_'+var, value)
        else:
            converted = auom(unit).convert(float(value), default_unit)
            setattr(self, '_'+var, converted)

    def __repr__(self):
        return f'<Transportation: {self.ID}>'

    def show(self):
        item = self.item
        impacts = self.impacts
        du = self.default_units
        info = f'Transportation: {self.ID}'
        info += f'\nImpact item   : {item.ID} [per trip]'
        info += f"\nLoad          : {f_num(self.load)} {du['load']}"
        info += f"\nDistance      : {f_num(self.distance)} {du['distance']}"
        info += f"\nInterval      : {f_num(self.interval)} {du['interval']}"
        info += f'\nTotal cost    : {f_num(self.cost)} {currency}'
        info += '\nTotal impacts :'
        print(info)
        if len(impacts) == 0:
            print(' None')
        else:
            index = pd.Index((i.ID+' ('+i.unit+')' for i in self.indicators))
            df = pd.DataFrame({
                'Impacts': tuple(self.impacts.values())
                },
                index=index)
            # print(' '*16+df.to_string().replace('\n', '\n'+' '*16))
            print(df.to_string())

    _ipython_display_ = show # funny that _ipython_display_ and _ipython_display behave differently

    def copy(self, new_ID=''):
        new = Transportation.__new__(Transportation)
        new.__init__(new_ID)
        new = copy_attr(new, self, skip=('_ID',))
        return new
    
    __copy__ = copy


    @property
    def item(self):
        '''[:class:`ImpactItem`] Item associated with this transportation activity.'''
        return self._item
    @item.setter
    def item(self, i):
        if not i:
            i = None
        elif isinstance(i, str):
            i = ImpactItem.get_item(i)
        elif not isinstance(i, ImpactItem):
            raise TypeError('Only `ImpactItem` or the ID of `ImpactItem` can be set, '
                            f'not {type(i).__name__}.')
        self._item = i

    @property
    def load_type(self):
        '''[str] Either "mass" or "volume".'''
        return self._load_type
    @load_type.setter
    def load_type(self, i):
        if i == 'mass':
            self.default_units['load'] = 'kg'
            self.default_units['quantity'] = 'kg*km'
        elif i == 'volume':
            self.default_units['load'] = 'm3'
            self.default_units['quantity'] = 'm3*km'
        else:
            raise ValueError('load_type can only be "mass" or "volume", '
                             f'not {i}.')
        self._load_type = i
        
    @property
    def indicators(self):
        ''' [tuple] Impact indicators associated with the transportation item.'''
        return self.item.indicators

    @property
    def load(self):
        '''
        [float] Transportation load each trip.

        .. note::
            
            Set this to 1 and let the load_unit match the functional unit of the item
            if load does not affect price and impacts.

        '''
        return self._load
    @load.setter
    def load(self, load, unit=''):
        self._update_value('load', load, unit)

    @property
    def distance(self):
        '''
        [float] Transportation distance each trip.

        .. note::
            
            Set this to 1 and let the `distance_unit` match the functional unit of the item
            if distance does not affect price and impacts.

        '''
        return self._distance
    @distance.setter
    def distance(self, distance, unit=''):
        self._update_value('distance', distance, unit)

    @property
    def quantity(self):
        '''[float] Quantity of item functional unit.'''
        quantity = auom(self.default_units['quantity']). \
            convert(self.load*self.distance, self.item.functional_unit)
        return quantity

    @property
    def interval(self):
        '''[float] Time between trips.'''
        return self._interval
    @interval.setter
    def interval(self, interval, unit=''):
        self._update_value('interval', interval, unit)

    @property
    def price(self):
        '''[float] Unit price of the item.'''
        return self.item.price

    @property
    def cost(self):
        '''[float] Total cost per trip.'''
        return self.price*self.quantity

    @property
    def impacts(self):
        '''[dict] Total impacts of this transportation item.'''
        impacts = {}
        for indicator, CF in self.item.CFs.items():
            impacts[indicator] = self.quantity*CF
        return impacts







